from abc import ABC, abstractmethod
from typing import Any, Optional

from era_5g_interface.dataclasses.control_command import ControlCommand


class TaskHandlerInitializationFailed(Exception):
    pass


class TaskHandler(ABC):
    """Abstract class.

    Task handler which takes care of receiving data from the NetApp
    client and passing them to the NetApp worker.
    """

    def __init__(self, sid: str, decoder=None) -> None:
        self.sid = sid
        self.websocket_id: Optional[str] = None
        self.frame_id = 0
        self.decoder = decoder
        self.last_timestamp = 0

    @abstractmethod
    def store_data(self, metadata: dict, data: Any) -> None:
        """This method is intended to pass the image to the worker (using
        internal queues, message broker or anything else).

        Args:
            metadata (dict): Arbitrary dictionary with metadata related to the image.
                The format is NetApp-specific.
            image (_type_): The image to be processed.
        """

        pass

    @abstractmethod
    def store_control_data(self, data: ControlCommand) -> None:
        """This method is intended to pass control commands to the worker.

        Args:
            data (ControlCommand): ControlCommand with control data.
        """

        pass

    @abstractmethod
    def clear_storage(self) -> None:
        """Clear storage used for communication with worker."""

        pass



from dataclasses import dataclass, field
from enum import IntEnum


class ControlCmdType(IntEnum):
    """Possible types of control commands for Network Application."""

    INIT = 1
    """Control Command used for initialization.

    Data required for initialization is provided using the data dict
    in the ControlCommand class.
    """

    RESET_STATE = 2
    """Reset current state of the Network Application."""

    GET_STATE = 3
    """Read and obtain current state of the Network Application."""

    SET_STATE = 4
    """Set newly defined state of the Network Application.

    The information about the new state can be provided using the
    data dict in the ControlCommand class.
    """

    SAVE_STATE = 5
    """Save current state on cloud storage."""

    LOAD_STATE = 6
    """Load saved state from cloud storage."""


@dataclass
class ControlCommand:
    """Dataclass containing information about control command for Network
    Application.

    Args:
        cmd_type (ControlCmdType): Type of the command to be sent.
        clear_queue (bool, optional): Clear previous unprocessed data (if any)
            in the internal queue. Default is False.
        data (dict, optional): Data to be passed along with the command,
            e.g. new state data in case of command with type
            ControlCmdType.SET_STATE.
    """

    cmd_type: ControlCmdType
    clear_queue: bool = False
    data: dict = field(default_factory=dict)



import logging
import time
from typing import Callable, Optional

logger = logging.getLogger("Rate timer")


class RateTimer:
    """Iteration timer used to control the speed of a loop.

    The timer takes into account the time consumed by processed work
    and sleeps only for the time remaining to complete the iteration.
    Code is loosely inpired by ROS Rate object and by
    https://stackoverflow.com/questions/23252796/control-the-speed-of-a-loop/23253617
    """

    def __init__(
        self,
        rate: Optional[float] = None,
        iteration_time: Optional[float] = None,
        name: str = "timer",
        time_function: Callable[[], float] = time.time,
        iteration_miss_warning: bool = False,
        interval_reinit_after_miss: bool = True,
        verbose: bool = False,
    ):
        """Initialize RateTimer.

        Args:
            rate (float, optional): Required number of iterations per
            second (i.e. rate in Hz, or FPS).
                Note: Either rate or iteration_time has to be provided.
            iteration_time (float, optional): Required length of one
                interation in seconds.
                Note: Either rate or iteration_time has to be provided.
            name (str, optional): User-defined name of the RateTimer
                object. Default is 'timer'.
            time_function (callable): Function that is called in order
                to determine current time. Default is time.time.
            iteration_miss_warning (bool): Logging flag. If True, each
                iteration misss is logged as a warning. Default is
                False.
            interval_reinit_after_miss (bool): Default is True.
                Flag to determine if a new interation interval should
                start after iteration miss (when set to True) or if the
                next iteration should be attempted as if there was no
                miss (when set to False).
            verbose (bool): Logging verbosity flag. Default is False.
        """

        if rate is None and iteration_time is None:
            raise RuntimeError("Either 'rate' or 'iteration_time' has to be specfied for RateTimer.")
        if rate is not None and iteration_time is not None:
            raise RuntimeError("Only 'rate' or 'interation_time' can be provided for RateTimer, but not both.")

        # Either rate in Hz (FPS) or iteration time (in seconds) can be given
        if iteration_time is not None:
            self.total_iteration_time = iteration_time
            self.rate = 1.0 / iteration_time
        elif rate is not None:
            self.total_iteration_time = 1.0 / rate
            self.rate = rate

        # time.time is chosen as default time_function,
        # but time.perf_counter could probably also be used
        self.time_function = time_function

        self.name = name
        self.iteration_miss_warning = iteration_miss_warning
        self.interval_reinit_after_miss = interval_reinit_after_miss
        self.verbose = verbose

        self.missed_iterations = 0
        self.times_called = 0
        self.creation_time = self.time_function()
        self.next_iteration_time = self.creation_time + self.total_iteration_time

        if self.verbose:
            logger.info(f"RateTimer ({self.name}) created.")

    def sleep(self) -> None:
        self.times_called += 1

        current_time = self.time_function()
        if current_time > self.next_iteration_time:
            # Processing took longer than defined iteration time. Do not sleep.
            self.missed_iterations += 1

            if self.interval_reinit_after_miss:
                # Plan new iterations starting from this time point (default behaviour)
                self.next_iteration_time = current_time + self.total_iteration_time
            else:
                # Another possibility is to try to keep the old interval between the iterations as before miss
                # (this may be problematic in case of larger time miss, because more iterations may then happen without sleep)
                self.next_iteration_time += self.total_iteration_time

            miss_msg = f"RateTimer ({self.name}): Iteration missed."
            if self.iteration_miss_warning:
                logger.warning(miss_msg)
            else:
                if self.verbose:
                    logger.info(miss_msg)

            # Do not perform any sleep
            return

        # time remaining to next iteration
        sleep_time = self.next_iteration_time - self.time_function()
        if sleep_time < 0:
            # sleep time must not be negative (it can happen due to threading)
            sleep_time = 0
        self.next_iteration_time += self.total_iteration_time
        time.sleep(sleep_time)

    def get_statistics(self):
        data = {
            "name": self.name,
            "rate": self.rate,
            "iteration_time": self.total_iteration_time,
            "times_called": self.times_called,
            "missed_iterations": self.missed_iterations,
        }
        return data


def rate_timer_example() -> None:
    """Example usage of RateTimer."""

    import random

    logger.setLevel(logging.INFO)

    rate = 2  # FPS
    iteration_time = 1.0 / rate
    logger.info(f"Using rate timer with rate {rate} Hz ({iteration_time} s).")

    # RateTimer should be created right before the loop starts
    rate_timer = RateTimer(rate, verbose=True, iteration_miss_warning=True)

    while True:
        logger.info(f"Time is: {time.time()}")

        # Generate random sleep time which is sometimes too long and forces iteration miss
        rand_time = random.uniform(iteration_time * 0.5, iteration_time * 1.1)
        time.sleep(rand_time)  # simulate work

        # Sleep until next interation
        rate_timer.sleep()


if __name__ == "__main__":
    rate_timer_example()



import logging

from av import Packet
from av.codec import CodecContext
from av.video.codeccontext import VideoCodecContext

# TODO: only for testing purpose
# Path("output").mkdir(parents=True, exist_ok=True)

logger = logging.getLogger("H264 decoder")


class H264Decoder:
    def __init__(self, fps: float, width: int, height: int) -> None:
        self.decoder: VideoCodecContext = CodecContext.create("h264", "r")
        self.decoder.width = width
        self.decoder.height = height
        self.decoder.framerate = fps
        self.decoder.pix_fmt = "yuv420p"

    def decode_packet_data(self, packet_data, format: str = "bgr24"):
        packet = Packet(packet_data)
        # TODO: only for testing purpose
        # logger.info(f"Decoding packet: {packet}")
        # Multiple frames? - This should not happen because on the encoders side one frame is always encoded and sent
        for frame in self.decoder.decode(packet):
            # TODO: only for testing purpose
            # logger.info(f"Frame {frame} with id {frame.index} decoded from packet: {packet}")
            # logger.info(f"frame.pts: {frame.pts}, frame.dts: {frame.dts}, frame.index: {frame.index}, "
            #            f"frame.key_frame: {frame.key_frame}, frame.is_corrupt: {frame.is_corrupt}, "
            #            f"frame.time: {frame.time}")
            # frame.to_image().save('output/frame-%04d.jpg' % frame.index)
            return frame.to_ndarray(format=format)



import logging
from enum import IntEnum
from queue import Empty, Full, Queue
from threading import Lock
from typing import Any, Optional

from era_5g_interface.dataclasses.control_command import ControlCommand
from era_5g_interface.task_handler import TaskHandler

logger: logging.Logger = logging.getLogger(__name__)


class QueueFullAction(IntEnum):
    DISCARD_OLDEST = 1
    DISCARD_NEW = 2
    RAISE = 3


class QueueDiscardError(Exception):
    """Exception which is raised when a problem occurs during discarding of old
    data when internal queue is full."""

    pass


class TaskHandlerInternalQ(TaskHandler):
    """Task handler which takes care of passing the data to the python internal
    queue for future processing.

    It could either be inherited to implement the _run method and read
    the data from any source or used directly and call the store_data
    method externally.
    """

    def __init__(
        self,
        sid: str,
        data_queue: Queue,
        if_queue_full: QueueFullAction = QueueFullAction.DISCARD_OLDEST,
        queue_put_lock: Optional[Lock] = None,
        **kw,
    ) -> None:
        """
        Constructor
        Args:
            sid (str): The session id obtained from NetApp client. It is used
                to match the results with the data sender.
            data_queue (Queue): The queue where the data and metadata should
                be passed to.
            if_queue_full (QueueFullAction, Optional):
                This parameter defines what should happen in case the internal
                queue is full and new data is to be added. Possible actions
                are defined by QueueFullAction Enum: DISCARD_OLDEST,
                DISCARD_NEW and RAISE.
                Default is DISCARD_OLDEST, which removes the oldest item
                from the queue to make up space for the new one. Option
                DISCARD_NEW throws away the newly provided data and keeps
                the queue intact. Choosing RAISE results in raising
                RuntimeError exception.
                Note that even DISCARD_OLDEST can result in discarding the
                new data - this happens in case the oldest data is a Control
                Command. Note also that, since Control Commands are considered
                more important than regular data, attempting to discard
                new Control Commad raises an exception, even in case
                the value is set to DISCARD_NEW.
            queue_put_lock (Lock, Optional):
                Lock for preventing multiple task handlers (or calls to
                store_data from multiple threads) from adding new data during
                the possible discarding of old data when queue is full. Default
                is None (the lock will be created internally.)
        """

        super().__init__(sid=sid, **kw)
        self._q: Queue = data_queue
        assert isinstance(if_queue_full, QueueFullAction)
        self.if_queue_full = if_queue_full
        if queue_put_lock is not None:
            self._q_put_lock = queue_put_lock
        else:
            self._q_put_lock = Lock()

    def data_queue_size(self) -> int:
        return self._q.qsize()

    def data_queue_occupancy(self) -> float:
        return self._q.qsize() / self._q.maxsize

    def store_data(self, metadata: dict, data: Any) -> None:
        """Method which will store the data (e.g. an image) to the queue for
        processing.

        Args:
            metadata (dict): Arbitrary dictionary with metadata related to the data.
                The format is NetApp-specific.
            data (Any): The data to be processed.

        Raises:
            QueueDiscardError: Raised when a problem occurs during discarding
                of old data when internal queue is full.
        """

        # make sure multiple calls of store_data do not interfere with each other
        with self._q_put_lock:
            self.frame_id += 1
            # logger.info(f"TaskHandlerInternalQ received frame id: {self.frame_id} with timestamp: {metadata['timestamp']}")

            try:
                self._q.put((metadata, data), block=False)
            except Full:
                if self.if_queue_full == QueueFullAction.DISCARD_NEW:
                    pass  # do nothing (discard the new data)
                elif self.if_queue_full == QueueFullAction.DISCARD_OLDEST:
                    # remove the oldest data from queue to make up space for the new one,
                    # but make sure not to discard control commands
                    space_available = False
                    with self._q.mutex:  # acquire the queue lock to prevent it from mutating during the removal
                        if isinstance(self._q.queue[0], ControlCommand):
                            logger.warning(
                                "Discarding new data instead of the old one to prevent losing a control command in queue."
                            )
                            # this situation could also raise an exception, if desired
                        else:
                            self._q._get()  # remove the old data from queue
                            space_available = True

                    if space_available:
                        self._q.put((metadata, data), block=False)
                else:
                    raise QueueDiscardError("Internal queue is full and no data was allowed to be discarded.")

    def store_control_data(self, data: ControlCommand) -> None:
        """Pass control commands to the worker using internal queue.

        Args:
            data (ControlCommand): ControlCommand with control data.

        Raises:
            QueueDiscardError: Raised when a problem occurs during discarding
                of old data when internal queue is full.
        """

        with self._q_put_lock:
            try:
                self._q.put(data, block=False)
            except Full:
                if self.if_queue_full == QueueFullAction.DISCARD_OLDEST:
                    with self._q.mutex:
                        if isinstance(self._q.queue[0], ControlCommand):
                            raise QueueDiscardError(
                                "Internal queue is full, and the new Control Command would discard a previous one."
                            )
                        else:
                            self._q._get()  # remove the old data from queue

                    self._q.put(data, block=False)
                else:
                    raise QueueDiscardError("Internal queue is full, Control Command cannot be processed.")

    def clear_storage(self) -> None:
        """Clear all items from internal queue."""

        while not self._q.empty():
            try:
                self._q.get(block=False)
            except Empty:
                break
            self._q.task_done()



import logging

import av
import numpy as np
from av.codec import CodecContext
from av.error import FFmpegError
from av.video.codeccontext import VideoCodecContext


class H264EncoderError(FFmpegError):
    """FFmpegError Exception."""

    pass


# TODO: only for testing purpose
# Path("input").mkdir(parents=True, exist_ok=True)

logger = logging.getLogger("H264 encoder")


class H264Encoder:
    def __init__(self, fps: float, width: int, height: int) -> None:
        self.encoder: VideoCodecContext = CodecContext.create("h264", "w")
        self.encoder.width = width
        self.encoder.height = height
        self.encoder.framerate = fps
        self.encoder.pix_fmt = "yuv420p"
        # TODO: only for testing purpose
        # self.encoder.options = {"crf": "0", "preset": "ultrafast", "tune": "zerolatency"}
        # self.encoder.options = {"preset": "ultrafast", "tune": "zerolatency","x264-params": "keyint=5"}
        # self.frame_id = 0
        self.encoder.options = {"preset": "ultrafast", "tune": "zerolatency"}

    def encode_ndarray(self, frame_data: np.ndarray, format: str = "bgr24"):
        frame = av.VideoFrame.from_ndarray(frame_data, format=format)
        # TODO: only for testing purpose
        # frame.to_image().save('input/frame-%04d.jpg' % self.frame_id)
        # self.frame_id += 1
        packets = []
        for packet in self.encoder.encode(frame):
            # TODO: only for testing purpose
            # logger.info(f"Frame {frame} encoded to packet: {packet}")
            packets.append(bytes(packet))
        if len(packets) > 1:
            logger.info(f"Frame {frame} encoded to multiple packets: {packets}")
        return b"".join(packets)



import logging
import math
import os
from typing import Any

import numpy as np
import requests
from requests.adapters import HTTPAdapter, Retry

logger = logging.getLogger(__name__)

NETAPP_ID = str(os.getenv("NETAPP_ID", "00000000-0000-0000-0000-000000000000"))
NETAPP_ID_ROS = NETAPP_ID.replace("-", "_")
MIDDLEWARE_ADDRESS = str(os.getenv("MIDDLEWARE_ADDRESS", "http://localhost"))
MIDDLEWARE_REPORT_INTERVAL = float(os.getenv("MIDDLEWARE_REPORT_INTERVAL", 1))
MAX_LATENCY = float(os.getenv("NETAPP_MAX_LATENCY", 100))


class LatencyMeasurements:
    """Class for holding data about processing times (latency)"""

    def __init__(self, num_latencies_to_keep: int = 10) -> None:
        self.num_latencies_to_keep = num_latencies_to_keep
        self.processing_latencies = np.zeros(num_latencies_to_keep)

    def store_latency(self, latency: float) -> None:
        # Remove the oldest entry and add the new one
        # Using strategy "Copy one before and substitute at the end" (fastest)
        self.processing_latencies[0:-1] = self.processing_latencies[1:]
        self.processing_latencies[-1] = latency

    def get_latencies(self) -> Any:
        return self.processing_latencies.tolist()

    def get_avg_latency(self) -> float:
        return float(np.mean(self.processing_latencies))


class HeartBeatSender:
    def __init__(self) -> None:
        self.retries = Retry(total=0, read=0, connect=0, backoff_factor=0, status_forcelist=[429, 500, 502, 503, 504])

        self.session: requests.Session = requests.Session()
        self.adapter: HTTPAdapter = HTTPAdapter(max_retries=self.retries)
        self.session.mount(MIDDLEWARE_ADDRESS, self.adapter)
        self.connection_error = False

    def send_middleware_heart_beat_request(self, headers: dict, json: dict) -> None:
        if not self.connection_error:
            logger.info(f"Sending heart beat to middleware: {json}")
            try:
                response = self.session.post(MIDDLEWARE_ADDRESS, headers=headers, json=json, timeout=(0.2, 0.2))
                if response.ok:
                    logger.debug(f"Middleware heart_beat response: {response.text}")
                else:
                    logger.warning(f"Middleware heart_beat response: {response.text}")
                    self.connection_error = True
            except requests.RequestException as e:
                logger.warning(f"Failed to connect to the middleware address: {MIDDLEWARE_ADDRESS}, {repr(e)}")
                self.connection_error = True

    def send_middleware_heart_beat(
        self, avg_latency: float, queue_size: int, queue_occupancy: float, current_robot_count: int
    ) -> None:
        # Latency can change over time, so reporting just the simple que occupancy can be misleading.
        # Instead, it is better to report occupancy in terms of:
        #  'total time estimated to be needed to process everything in the queue' / 'required max latency'
        processing_time_occupancy = queue_size * avg_latency / MAX_LATENCY

        if queue_size == 0:
            # If the queue is empty then no estimate can be made about robot count limit,
            # but most likely at least one more robot can be added.
            hard_robot_count_limit = current_robot_count + 1
            optimal_robot_count_limit = current_robot_count + 1
        elif avg_latency == 0:
            # If there are no latency measurements, the maximum number of robots cannot be estimated
            # using processing_time_occupancy, but we can still try to use queue_occupancy.
            hard_robot_count_limit = math.floor(current_robot_count / queue_occupancy)
            optimal_robot_count_limit = math.floor(hard_robot_count_limit * 0.8)
        else:
            hard_robot_count_limit = math.floor(current_robot_count / processing_time_occupancy)
            optimal_robot_count_limit = math.floor(hard_robot_count_limit * 0.8)

        data = {
            "Id": NETAPP_ID,
            "CurrentRobotsCount": current_robot_count,
            "OptimalLimit": optimal_robot_count_limit,
            "HardLimit": hard_robot_count_limit,
        }
        headers = {"Content-type": "application/json"}
        self.send_middleware_heart_beat_request(headers=headers, json=data)
        # TODO: only for testing purpose
        # logger.info(f"Heart beat published to middleware: {data}")
